/****************************************************************************
*
*  This module implements the code for Bezier bicubic patch shapes
*
*  This file was written origionally by Alexander Enzmann.  He wrote the code 
*  for bezier bicubic patches and generously provided povray with these 
*  enhancements.
*
*  This file was converted and edited for use in a povray include file by
*  Loial Raven.
*
*****************************************************************************/
/* bicubic_line.inc
 this is an exact conversion from the .c file, except it makes a line instead
 of a surface. 

If you are not using megapov, becubic_line itself doesn't work, but the 
bezier_value works fine.

the use is: (if you have use_megapov declared)
  bicubic_line (points, rad)
where points is an array of four points and the rad is the radius of the 
spheresweep

the bezier curve created by this touches either end point and are weighted
directly towards the two middle points.

there is also a quality modifier, all you have to do is #declare it before you 
use the macro. the number is the number of segments in the sweep.
you can define it like this:
  #declare bicubic_line_quality = 20;
for a 20 segment line

just as a note, the bezier_value macro seems to work fine with 2d points as well as 3d
*/

#ifdef (use_megapov)
  #version unofficial MegaPov 0.5;
#end
         
#ifndef (__bicubic_line_inc)
	#declare __bicubic_line_inc = 1;

	#declare C = array[4] { 1.0, 3.0, 3.0, 1.0 }
                                            
  #ifndef (vcross3)                                          
	  #macro vcross3 (point1, basepoint, point2)
		  vcross(point1-basepoint, point2-basepoint)
  	#end
	#end
	
	#macro bezier_value(points, u0)
		#local P = <0,0>;
		#local i = 0;
		#local c = 0.0;
		#local _t = 0.0; #local ut = 0.0;
		#local _u = array[4]  { 0.0, 0.0, 0.0, 0.0 }
		#local uu  = array[4] { 0.0, 0.0, 0.0, 0.0 }
		#local du = array[4]  { 0.0, 0.0, 0.0, 0.0 }
		#local duu = array[4] { 0.0, 0.0, 0.0, 0.0 }
		#local U1 = <0,0,0>;
		
		/* Calculate binomial coefficients times coordinate positions. */
		
		#local _u[0] = 1.0; 
		#local uu[0] = 1.0; 
		#local du[0] = 0.0; 
		#local duu[0] = 0.0;
		
		#local i = 1;
		#while (i < 4)
			#local _u[i] = _u[i - 1] * u0;  
			#local uu[i] = uu[i - 1] * (1.0 - u0);
			
			#local du[i] = i * _u[i - 1];  
			#local duu[i] = -i * uu[i - 1];
			#local i = i + 1;
		#end
		
		/* Now evaluate position and tangents based on control points. */
			
		#local i = 0;
		#while (i < 4)
			#local c = C[i];
			
			#local ut = _u[i] * uu[3 - i];
			
			#local _t = c * ut;
			
			#local P = P + _t*points[i];
			
			#local _t = c * (du[i] * uu[3 - i] + _u[i] * duu[3 - i]);
			
			#local U1 = U1 + _t*points[i];
			
			#local i = i + 1;
		#end
		P
	#end
	
#ifdef (use_megapov)
	#macro bicubic_line (points, rad)
		#ifndef (bicubic_line_quality) #declare bicubic_line_quality = 20; #end
		sphere_sweep {
			linear_sphere_sweep
			#local step = 1/bicubic_line_quality;
			1/step+1
			#local i = 0.0;
			#while (i <= 1)
				bezier_value(points,i), rad
				#local i = i + step;
			#end
			points[3], rad
			sphere_sweep_depth_tolerance 1.0e-3
		}	
	#end
#end
		
	#macro bezier_normal (points, u0)
		#if (u0 +0.0001 <= 1)
			vcross3 (<0,0,0>+bezier_value(points, u0+0.001), <0,0,0>+bezier_value(points, u0), <0,0,0>+bezier_value(points, u0)+y*0.001)
		#else
			vcross3 (<0,0,0>+bezier_value(points, u0), <0,0,0>+bezier_value(points, u0-0.001), <0,0,0>+bezier_value(points, u0-0.001)+y*0.001)
		#end
	#end

#end
